/********************************************************************************/
/*    H   H Y   Y TTTTT EEEEE  CCC       HYTEC ELECTRONICS LTD                  */
/*    H   H  Y Y    T   E     C          5 Cradock Road,                        */
/*    HHHHH   Y     T   EEE   C          Reading, Berks.    Tel: 0118 9757770	*/
/*    H   H   Y     T   E     C          RG2 0JT            Fax: 0118 9757566	*/
/*    H   H   Y     T   EEEEE  CCC       Web: www.hytec-electronics.co.uk       */
/********************************************************************************/
/********************************************************************************/
/*      ___________________________________________________________________     */
/*      | H Y T E C   F i r m w a r e  U p g r a d e  A s y n  D r i v e r |    */
/*           ---------------------------------------------------------------    */
/*                                                                              */
/*  Source file name      :-      drvFirmware.c                                 */
/*                                                                              */
/*  Initial creation date :-      17-Jul-2013                                   */
/*                                                                              */
/*  Original Developers   :-      Jim Chen						                */
/*                                                                              */
/********************************************************************************/
/*                                                                              */
/*  Description :- This code is for Hytec firmware upgradeable IP module to     */
/*  upgrade firmware without JTag programmer. 				                    */
/*                                                                              */
/*  (C)2013 Hytec Electronics Ltd.                                              */
/*                                                                              */
/********************************************************************************/
/*                                                                              */
/* Modification record :-                                                       */
/*                                                                              */
/*  Date modified |  Person   | Version | Description of modification           */
/* ---------------|-----------|---------|-------------------------------------- */
/*    17/07/13    |  JC		  | 1.0     | Initial creation with new asyn model  */    
/*    04/09/13    |  JC		  | 1.1     | Now works for both VME and IOC9010.	*/    
/*                |           |         |    		                            */    
/********************************************************************************/

#define DRIVER_VERSION	101

/******************/
/* Hytec Module includes */
/******************/
#include "drvFirmware.h"


/********************************************************************/
/*														           */
/* Public Sector												  */
/*													             */
/****************************************************************/
/* Constructor:
 * Where: -
 *     portName = asyn port name
 *   carrierNum = carrier index number when adding a carrier card by ipacAddHy8002 call
 *    ipSlotNum = IP slot number on the carrier board (0 = A, 1 = B, 2 = C, 3 = D, 4 = E, 5 = F),
 *     fileName = firmware filename to upload. Ignored in mode 0 and 1.
 *     	   mode = 0:Memory test
 * 				  1:View firmware 
 *			  	  2:Read PROM and compare with specified .bit file, 
 * 				  3:Firmware upgrade
 */    
HyFWUpgradeAsyn::HyFWUpgradeAsyn(char *portName, int carrierNum, int ipSlotNum, char *fileName, int mode)    
		: asynPortDriver(portName, MAX_ADC_CHANNELS, NUM_PARAMS, 
        asynInt32Mask | asynDrvUserMask, asynInt32Mask, 
        ASYN_MULTIDEVICE, 1, /* ASYN_CANBLOCK=0, ASYN_MULTIDEVICE=1, autoConnect=1 */
        0, 0)  /* Default priority and stack size */

{
    const char *hytecstr = "(Hytec Electronics Ltd., Reading, UK)";
    int strok, counter;
  	char lstr[7];
  	USHORT idmem;
    int ishytec;
	int ver;

	/* sweep data channels definition */
    createParam(hyFWUpgradeErrorString, 		asynParamInt32, 	&hyFWUpgradeError_);
    createParam(hyFWUpgradeDriverVersionString,	asynParamInt32, 	&hyFWUpgradeDriverVersion_);
	
	/* Save the parameters in the driver structure    */
    this->carrierNum		= carrierNum;
    this->ipSlotNum			= ipSlotNum;
    this->mode			 	= mode;

	setIntegerParam(hyFWUpgradeDriverVersion_, DRIVER_VERSION);

	/* Map Register address spaces  */
	this->ipio = (IO_MEMORY *) ipmBaseAddr(carrierNum, ipSlotNum, ipac_addrIO);
    if (this->ipio == NULL)
    {  
        /* can't register card register address space */
    	printf("configHyFWUpgrade: Unable to register IO space\n");
        return;
    }
    
    /* map the memory address spaces  */
    this->ipmem = (USHORT *) ipmBaseAddr(carrierNum, ipSlotNum, ipac_addrMem);
    if (this->ipmem == NULL)
    {  
        /* can't register card Memory address space */
     	printf("configHyFWUpgrade: Unable to register memory space\n");
        return;
    }
     
    /* Compare the HYFWUpgrade prom header, manufacturer's ID and card model
     * number with the expected values. */
  	idmem =  *((USHORT*)this->ipio->prom.vita4);
  	if((idmem == 0x5649) || (idmem == 0x4956));
  	{
  		lstr[0] = 0x56; lstr[1] = 0x49;
  	}
  	idmem =  *((USHORT*)this->ipio->prom.vita4 + 1);
  	if((idmem == 0x5441) || (idmem == 0x4154));
  	{	
  		lstr[2] = 0x54; lstr[3] = 0x41;
  	}
  	idmem =  *((USHORT*)this->ipio->prom.vita4 + 2);
  	if((idmem == 0x3420) || (idmem == 0x2034));
  	{	
  		lstr[4] = 0x20; lstr[5] = 0x34;
  	}
  	lstr[6] = 0;

    strok = memcmp(lstr, VITA4, sizeof(this->ipio->prom.vita4)) == 0;
    ishytec = this->ipio->prom.hytecIdHigh == HYTECID_HI  &&
                  this->ipio->prom.hytecIdLow == HYTECID_LO;

	/* check if we have been asked to be more chatty */
    /* start with some basic information */
    printf("\nPROM: Contents at offset - %#08x\n", (int)&this->ipio->prom);
    printf("PROM: Prom header string - '%6s'\n", lstr);
    printf("PROM: Manufacturer's ID  - 0x%04x%04x %s\n",
        (int)this->ipio->prom.hytecIdHigh, (int)this->ipio->prom.hytecIdLow, 
        ishytec ? hytecstr : "(Unknown)");
    printf("\n");
    /* output a bit more about this specific model and card */
    printf("PROM: Model #/revision   - %#04hx rev. %#04hx\n",
        (int)this->ipio->prom.model, (int)this->ipio->prom.revision);

	ver = (int)this->ipio->prom.revision;
	ver = ver/4096*1000 + (ver%4096)/256*100 + ((ver%4096)%256)/16*10 + ((ver%4096)%256)%16;

    printf("PROM: Serial number      - %hu\n", (int)this->ipio->prom.serialNumber);
    printf("PROM: Driver IDs         - %#04hx, %#04hx\n",
        (int)this->ipio->prom.driverIdHigh, (int)this->ipio->prom.driverIdLow);
    printf("PROM: Device flags       - %#04hx\n", (int)this->ipio->prom.flags);
    printf("PROM: Bytes used         - %#04hx (%hd)\n",
        (int)this->ipio->prom.byteCount, (int)this->ipio->prom.byteCount);
    printf("\n");

    /* check if everything passed okay */
    if (!strok)
        printf("PROM: INVALID PROM HEADER; EXPECTED '%s'\n", VITA4);
    if (!ishytec)
        printf("PROM: UNSUPPORTED MANUFACTURER ID; EXPECTED 0x%04x%04x, %s\n",
            HYTECID_HI, HYTECID_LO, hytecstr);

	switch(this->mode)
	{
	case 0:	//memory test
		printf("******* Memory Read/Write Test ******\n");
		error = memTest(1024*1024);
		printf("Test finished (read/write) with %d error(s).\n\n ", error);
		break;
	case 1:	//View firmware version
		break;
	case 2:	//read PROM
	case 21://read PROM
	case 22://read PROM
		printf("******* Verify Firmware ******\n");
	    if(checkFilename(fileName, this->ipio->prom.model) != 0)
	    {
	        printf("The .bit file (%s) doesn't match the model number %x. Check the IP card and the st.cmd file and try again!\n\n", fileName, this->ipio->prom.model);
			return;
	    }

	    clearBuffer();
		printf("Open file %s....", fileName);
		this->fileLength = openFile(fileName);
		if(this->fileLength <= 0)
		{
			printf("Unable to open file %s\n\n", fileName);
			return;
		}

		//Read back from the PROM
		readBack();
		if(this->mode == 2)
			error = compare(0, 0);			//print only symbol and no pause
		else if(this->mode == 21)
			error = compare(1, 0);			//print block with errors and no pause
		else
			error = compare(2, 0);			//print everything and no pause

		if(error != 0)			
			printf("\nFirmware doesn't match the .bit file. It has %d error(s).\n\n", error);
		else
			printf("\nVerified successfully with %d error.\n\n", error);
		break;
	case 3:	//firmware upgrade
	case 31:	//firmware upgrade
	case 32:	//firmware upgrade
		printf("******* Firmware Upgrade ******\n");
	    if(checkFilename(fileName, this->ipio->prom.model) != 0)
	    {
	        printf("The .bit file (%s) doesn't match the model number %x. Check the IP card and the st.cmd file and try again!\n\n", fileName, this->ipio->prom.model);
			return;
	    }

		clearBuffer();
		//Open file
		printf("Open file %s....", fileName);
		this->fileLength = openFile(fileName);
		if(this->fileLength <= 0)
		{
			printf("Unable to open file %s\n\n", fileName);
			return;
		}

		//Check if it is busy first
		this->ipio->reg.extStatus = 0xC00;
		printf("Check if it is busy (%x)... ", this->ipmem[0]);
		counter = 10000;
		while(((USHORT)(this->ipmem)[0] & 0x80) == 0) 
		{
			counter--;
			if(counter <= 0) break;
		}
		if(counter <= 0)
		{
			printf("SPI is busy, program times out. Please quit ioc and start again!\n\n");
			return;
		}
		printf("ok!\n");


		//flood to the memory
		printf("Flash to memory (%x)......\n", this->ipmem[0]);
		error = writeFlash();
		if(error < 0)
		{
			printf("SPI is busy, program times out. Please quit ioc and start again!\n\n");
			return;
		}
		else
			printf("done!\n");
	

		//Read back from the PROM
		readBack();
		//verify write and read
		if(this->mode == 3)
			error = compare(0, 0);			//print only symbol and no pause
		else if(this->mode == 31)
			error = compare(1, 0);			//print block with errors and no pause
		else
			error = compare(2, 0);			//print everything and no pause
	
		if(error == 0) 
		{
			printf("\nNo error! Please power cycle the IOC and use mode 1 to check the new firmware.\n\n");
			return;
		}
		printf("\nThere are %d errors. Please don't power off the IOC. Just exit the IOC and re-run the upgrade. If still errors after 3 times, please contact Hytec support.\n\n", error);
		break;
	}
}


asynStatus HyFWUpgradeAsyn::readInt32(asynUser *pasynUser, epicsInt32 *value)
{
	asynStatus status = asynSuccess;
	epicsInt32 channel;
	int function = pasynUser->reason;
   	
	this->getAddress(pasynUser, &channel);
		
    if(function == hyFWUpgradeError_)
	{
		*value = error > 0 ? 1 : 0;
	}
	else
		asynPortDriver::readInt32(pasynUser, value);
	
  	callParamCallbacks();
	return( status );
}


//*************private section*********************
//
//

/* Function: int HyFWUpgradeAsyn::clearBuffer()
 * Description: Clear the buffers
 * return: OK
 */
int HyFWUpgradeAsyn::clearBuffer()
{
	int i;
	//clear buffers
	for(i=0; i<BANKSIZE; i++)
	{
		this->buf[i] = 0;
		if(i < BANKSIZE/2)
		{
			this->wordbuf[i] = 0;
			this->verifyBuf[i] = 0;
		}
	}
	return 0;
}

/* Function: int HyFWUpgradeAsyn::openFile(char * fileName)
 * Description: Open the .bit file
 * Where: -
 *     fileName = .bit firmware filename 
 * return: length of the file in byte 
 */
int HyFWUpgradeAsyn::openFile(char * fileName)
{
	int i;
	FILE *file;
	unsigned long fileLen;

	file = fopen(fileName, "rb");
	if (!file)
		return -1;
	
	//Get file length
	fseek(file, 0, SEEK_END);
	fileLen=ftell(file);
	fseek(file, 0, SEEK_SET);
	printf("File length = %d\n", (int)fileLen);

	//Read file contents into buffer
	printf("Read file into buffer....");
	fread(this->buf, fileLen, 1, file);
	printf("ok.\n");

	fclose(file);
	printf("File closed.\n\n");

	//convert byte array to word array
	for(i=0; i<(int)fileLen; i=i+2)
	{
		this->wordbuf[i/2] = ((USHORT)this->buf[i] << 8) + this->buf[i+1];
	}

	return fileLen;
}


/* Function: int HyFWUpgradeAsyn::checkFilename(char * fileName, int model)
 * Description: verify if the file name is correct for the specified model
 * 				by checking the filename that includes the model number
 * Where: -
 *     fileName = .bit file name 
 *   	model 	= model number. e.g. 8424, 8522 etc
 * return: 0=correct, -1=incorrect. 
 */
int HyFWUpgradeAsyn::checkFilename(char * fileName, int model)
{
	char mod[16], *pch;
	sprintf(mod,"%x", model);  
    pch = strstr(fileName, mod);
	if(pch)	return 0;
	else return -1;
}

/* Function: int HyFWUpgradeAsyn::writeFlash()
 * Description: Write the firmware bit file to the flash
 * return: OK
 */
int HyFWUpgradeAsyn::writeFlash()
{
	int i, j, k = 45, n, counter;
	for(j=0; j<567; j++)							//567 times
	{
		this->ipio->reg.extStatus = 0x0800;			//enable flash access
		for(i=0; i<132; i++)						//each time 132 words
		{
			(this->ipmem)[i] = this->wordbuf[k];
			k++;
		}
		this->ipio->reg.extStatus = 0x0C00;			//set up the command
		(this->ipmem)[(j<<9)/2] = (USHORT) 0;		//do the erase and write
			
		for(n=0;n<100;n++) n = n;					//allow a little bit settle time.....
		
		counter = 10000;
		while(((USHORT)(this->ipmem)[0] & 0x80) == 0) 
		{
			counter--;
			if(counter <= 0) break;
		}
		if(counter <= 0)
			return -1;
		
		printf(".");
		if((j > 0) && ((j + 1) % 50 == 0)) printf("\n");
	}
	this->ipio->reg.extStatus = 0x0000;				//disable flash access
	return 0;
}


/* Function: int readBack()
 * Description: Read PROM contents into verifyBuf for reabable IP card.
 * return: OK
 */
int HyFWUpgradeAsyn::readBack()
{
	int i, j, k = 45;

	this->ipio->reg.extStatus = 0x0800;				//enable flash access
	for(j=0; j<567; j++)							//567 times
	{
		for(i=0; i<132; i++)						//each time 132 words
		{
			this->verifyBuf[k] = (USHORT)(this->ipmem)[i+((j<<9)/2)];		//read back
			k++;
		}
	}
	this->ipio->reg.extStatus = 0x0000;				//disable flash access	
	return 0;
}


/* Function: int compare(int fullText, int pause)
 * Description: compare the .bit file contents to the PROM read back
 *				and print out the result if requested.
 * Where: -
 *     fullText = 0:print out a '.' char if ok or a '#' if not for a block, 
 * 				  1:print out the error block, 
 *				  2:print everything
 *   	pause 	= 0:no pause between the block prints, 
 *				  1:pause every block (only avial for Linux)
 * return: number of errors 
 */
int	HyFWUpgradeAsyn::compare(int fullText, int pause)
{
	int i = 0, j, k = 45, n = 0, error = 0;
	char buffer[10000], temp[64];

	printf("\nVerifying......\n");
	strcpy(buffer, "");
	for(j=0; j<567; j++)							//567 x 132 words
	{
		for(i=0;i<132;i++)
		{
			if((i % 12) == 0) 
			{
				sprintf(temp, "%04d: ", k-45);
				strcat(buffer, temp);
			}
			if(this->verifyBuf[k]==this->wordbuf[k])
				sprintf(temp, "%04x[%04x] ", this->wordbuf[k], this->verifyBuf[k]);
			else
			{
				sprintf(temp, "%04x#%04x# ", this->wordbuf[k], this->verifyBuf[k]);
				n++;
				error++;
			}
		
			strcat(buffer, temp);
			if(((i+1) % 12) == 0) 
			{
				strcat(buffer, "\n");
			}
			k++;
		}
		if(fullText == 2)
			printf("%s\n", buffer);
		else if(fullText == 1)
		{
			if(n != 0) printf("%s\n", buffer);
		}
		else if(n != 0) 
			printf("#");
		else
			printf(".");
		
		if((j > 0) && ((j + 1) % 50 == 0)) printf("\n");
		n = 0;
		strcpy(buffer, "");
		
		if(pause) getchar();
	}
	return error;
}

/* Function: int HyFWUpgradeAsyn::memTest(int size)
 * Description: Memory read and write test
 * Where: -
 *     size = the size in WORD to test. Maximum 1024x1024 
 * return: number of errors
 */
int HyFWUpgradeAsyn::memTest(int size)
{
	int i, error=0;
	
	this->ipio->reg.status = 0;
	
	if(size > 1024*1024)
	{
		printf("Maximum memory size is 1024x1024.\n");
		return -1;
	}
	
	printf("Write 0x5555....\n");
	for(i=0; i<size; i++)							//size times
		(this->ipmem)[i] = 0x5555;					//write 0x5555

	printf("Read and Verify 0x5555....\n");
	for(i=0; i<size; i++)							//size times
	{
		if((USHORT)(this->ipmem)[i] != 0x5555)
			error++;			
	}
	
	printf("Write 0xAAAA....\n");
	for(i=0; i<size; i++)							//size times
		(this->ipmem)[i] = 0xAAAA;					//write 0xAAAA

	printf("Read and Verify 0xAAAA....\n");
	for(i=0; i<size; i++)							//size times
	{
		if((USHORT)(this->ipmem)[i] != 0xAAAA)
			error++;			
	}
	
	return error;	
}
